@@ -299,15 +299,49 @@ class Balloon:
         if len(self.window) > side:
             self.window[side].append_sstp_marker()
 
+    def append_link_in(self, side, label):
+        if self.synchronized:
+            for side in self.synchronized:
+                if len(self.window) > side:
+                    self.window[side].append_link_in(label)
+        else:
+            if len(self.window) > side:
+                self.window[side].append_link_in(label)
+
+    def append_link_out(self, side, label, value):
+        if self.synchronized:
+            for side in self.synchronized:
+                if len(self.window) > side:
+                    self.window[side].append_link_out(label, value)
+        else:
+            if len(self.window) > side:
+                self.window[side].append_link_out(label, value)
+
     def append_link(self, side, label, value, newline_required=0):
         if self.synchronized:
             for side in self.synchronized:
                 if len(self.window) > side:
-                    self.window[side].append_link(label, value,
-                                                  newline_required)
+                    self.window[side].append_link_in(label)
+                    self.window[side].append_text(value)
+                    self.window[side].append_link_out(label, value)
+                    if newline_required:
+                        self.window[side].set_newline()
         else:
             if len(self.window) > side:
-                self.window[side].append_link(label, value, newline_required)
+                self.window[side].append_link_in(label)
+                self.window[side].append_text(value)
+                self.window[side].append_link_out(label, value)
+                if newline_required:
+                    self.window[side].set_newline()
+
+    def append_meta(self, side, tag):
+        if self.synchronized:
+            for side in self.synchronized:
+                if len(self.window) > side:
+                    self.window[side].append_meta(tag)
+        else:
+            if len(self.window) > side:
+                self.window[side].append_meta(tag)
 
     def append_image(self, side, path, x, y):
         if len(self.window) > side:
@@ -518,6 +552,7 @@ class BalloonWindow:
             pango_size = self.sstp_font_desc.get_size()
             if pango_size == 0:
                 pango_size = size * 3 / 4 # convert from Windows to GTK+
+                pango_size *= pango.SCALE
             self.sstp_font_desc.set_size(pango_size * scale / 100)
             self.sstp_layout.set_font_description(self.sstp_font_desc)
             self.sstp_layout.set_wrap(pango.WRAP_CHAR)
@@ -769,6 +804,35 @@ class BalloonWindow:
         cr.set_source_pixbuf(self.balloon_pixbuf, 0, 0)
         cr.paint_with_alpha(self.__alpha_channel)
 
+    def set_markup(self, index, text):
+        tags_ = ('sup', 'sub', 's', 'u')
+        count_ = {}
+        for tag_ in tags_:
+            count_[tag_] = 0
+        markup_list = []
+        for sl, sn, tag in self.meta_buffer:
+            if sl == index:
+                markup_list.append((sn, tag))
+        markup_list.sort()
+        markup_list.reverse()
+        for sn, tag in markup_list:
+            text = ''.join((text[:sn], tag, text[sn:]))
+            if tag[1] == '/':
+                tag_ = tag[2:-1]
+                assert tag_ in tags_
+                count_[tag_] -= 1
+                if count_[tag_] < 0:
+                    text = ''.join(('<', tag_, '>', text))
+                    count_[tag_] += 1
+            else:
+                tag_ = tag[1:-1]
+                assert tag_ in tags_
+                count_[tag_] += 1
+                if count_[tag_] > 0:
+                    text = ''.join((text, '</', tag_, '>'))
+                    count_[tag_] -= 1
+        return text
+
     def redraw(self, darea=None, event=None):
         if not self.__shown:
             return True
@@ -820,11 +884,13 @@ class BalloonWindow:
             if self.text_buffer[i].endswith('\n[half]'):
                 new_y = int(y + (self.font_height + self.line_space) / 2)
                 self.update_line_regions(line + 1, new_y)
-                self.layout.set_text(self.text_buffer[i][:-7])
+                markup = self.set_markup(i, self.text_buffer[i][:-7])
+                self.layout.set_markup(markup)
                 darea.window.draw_layout(self.text_gc[gtk.STATE_NORMAL],
                                          x, y, self.layout)
             else:
-                self.layout.set_text(self.text_buffer[i])
+                markup = self.set_markup(i, self.text_buffer[i])
+                self.layout.set_markup(markup)
                 darea.window.draw_layout(self.text_gc[gtk.STATE_NORMAL],
                                          x, y, self.layout)
             for l, c in self.sstp_marker:
@@ -859,36 +925,43 @@ class BalloonWindow:
                     break
                 x, y, w, h = self.line_regions[n - self.lineno]
                 if sl == el:
-                    self.layout.set_text(self.text_buffer[n][:sn])
+                    markup = self.set_markup(n, self.text_buffer[n][:sn])
+                    self.layout.set_markup(markup)
                     text_w, text_h =  self.layout.get_pixel_size()
                     x += text_w
-                    self.layout.set_text(self.text_buffer[n][sn:en])
+                    markup = self.set_markup(n, self.text_buffer[n][sn:en])
+                    self.layout.set_markup(markup)
                     text_w, text_h =  self.layout.get_pixel_size()
                     w = text_w
                     start = sn
                     end = en
                 elif n == sl:
-                    self.layout.set_text(self.text_buffer[n][:sn])
+                    markup = self.set_markup(n, self.text_buffer[n][:sn])
+                    self.layout.set_markup(markup)
                     text_w, text_h =  self.layout.get_pixel_size()
                     x += text_w
-                    self.layout.set_text(self.text_buffer[n][sn:])
+                    markup = self.set_markup(n, self.text_buffer[n][sn:])
+                    self.layout.set_markup(markup)
                     text_w, text_h =  self.layout.get_pixel_size()
                     w = text_w
                     start = sn
                     end = len(self.text_buffer[n])
                 elif n == el:
-                    self.layout.set_text(self.text_buffer[n][:en])
+                    markup = self.set_markup(n, self.text_buffer[n][:en])
+                    self.layout.set_markup(markup)
                     text_w, text_h =  self.layout.get_pixel_size()
                     w = text_w
                     start = 0
                     end = en
                 else:
-                    self.layout.set_text(self.text_buffer[n])
+                    markup = self.set_markup(n, self.text_buffer[n])
+                    self.layout.set_markup(markup)
                     text_w, text_h =  self.layout.get_pixel_size()
                     w = text_w
                     start = 0
                     end = len(self.text_buffer[n])
-                self.layout.set_text(self.text_buffer[n][start:end])
+                markup = self.set_markup(n, self.text_buffer[n][start:end])
+                self.layout.set_markup(markup)
                 self.redraw_area(x, y, w, h)
                 if not clear:
                     darea.window.draw_rectangle(self.cursor_gc, True,
@@ -912,15 +985,19 @@ class BalloonWindow:
                         break
                     x, y, w, h = self.line_regions[n - self.lineno]
                     if n == sl:
-                        self.layout.set_text(self.text_buffer[n][:sn])
+                        markup = self.set_markup(n, self.text_buffer[n][:sn])
+                        self.layout.set_markup(markup)
                         text_w, text_h =  self.layout.get_pixel_size()
                         x += text_w
                     if n == sl and n == el:
-                        self.layout.set_text(self.text_buffer[n][sn:en])
+                        markup = self.set_markup(n, self.text_buffer[n][sn:en])
+                        self.layout.set_markup(markup)
                     elif n == el:
-                        self.layout.set_text(self.text_buffer[n][:en])
+                        markup = self.set_markup(n, self.text_buffer[n][:en])
+                        self.layout.set_markup(markup)
                     else:
-                        self.layout.set_text(self.text_buffer[n])
+                        markup = self.set_markup(n, self.text_buffer[n])
+                        self.layout.set_markup(markup)
                     text_w, text_h =  self.layout.get_pixel_size()
                     w = text_w
                     if x <= px < x + w and y <= py < y + h:
@@ -1016,6 +1093,7 @@ class BalloonWindow:
         self.selection = None
         self.lineno = 0
         self.text_buffer = []
+        self.meta_buffer = []
         self.link_buffer = []
         self.newline_required = 0
         self.images = []
@@ -1028,12 +1106,21 @@ class BalloonWindow:
     def reset_text_count(self):
         self.text_count = 0
 
+    def set_newline(self):
+        self.newline_required = 1
+
     def append_text(self, text):
-        if not self.text_buffer or self.newline_required:
+        if not self.text_buffer:
+            s = ''
+            column = 0
+            index = 0
+        elif self.newline_required:
             s = ''
             column = 0
             self.newline_required = 0
+            index = len(self.text_buffer)
         else:
+            index = len(self.text_buffer) - 1
             s = self.text_buffer.pop(-1)
             column = len(s)
         i = len(s)
@@ -1059,7 +1146,8 @@ class BalloonWindow:
             n = i + 1
             if not self.__shown:
                 self.show()
-            self.layout.set_text(text[p:n])
+            markup = self.set_markup(index, text[p:n])
+            self.layout.set_markup(markup)
             text_width, text_height =  self.layout.get_pixel_size()
             if text_width > self.line_width:
                 self.text_buffer.append(text[p:i])
@@ -1094,50 +1182,43 @@ class BalloonWindow:
         self.append_text(space)
         self.draw_last_line(offset)
 
-    def append_link(self, link_id, text, newline_required=0):
-        if not text:
-            return
-        self.show()
-        raw_text = text
-        if not self.text_buffer or newline_required:
-            offset = ''
-            sl = len(self.text_buffer)
+    def append_link_in(self, link_id):
+        if not self.text_buffer:
+            sl = 0
             sn = 0
         else:
-            offset = self.text_buffer[-1]
             sl = len(self.text_buffer) - 1
             sn = len(self.text_buffer[-1])
-        start = i = 0
-        line_number = 0
-        while 1:
-            n = i + 1
-            self.layout.set_text(''.join((offset, text[start:n])))
-            text_w, text_h = self.layout.get_pixel_size()
-            if text_w > self.line_width:
-                if line_number == 0 and \
-                   not newline_required and self.text_buffer:
-                    self.text_buffer[-1] = ''.join(
-                        (self.text_buffer[-1], text[start:i]))
-                    self.draw_last_line()
-                else:
-                    self.text_buffer.append(text[start:i])
-                line_number = line_number + 1
-                start = i
-                offset = ''
-            i = n
-            if i >= len(text):
-                if line_number == 0 and \
-                   not newline_required and self.text_buffer:
-                    self.text_buffer[-1] = ''.join(
-                        (self.text_buffer[-1], text[start:i]))
-                else:
-                    self.text_buffer.append(text[start:i])
+        self.link_buffer.append((sl, sn, sl, sn, link_id, '', ''))
+
+    def append_link_out(self, link_id, text):
+        if not text:
+            return
+        raw_text = text
+        if not self.text_buffer:
+            el = 0
+            en = 0
+        else:
+            el = len(self.text_buffer) - 1
+            en = len(self.text_buffer[-1])
+        for i in range(len(self.link_buffer)):
+            if self.link_buffer[i][4] == link_id:
+                sl = self.link_buffer[i][0]
+                sn = self.link_buffer[i][1]
+                self.link_buffer.pop(i)
+                self.link_buffer.insert(i, (sl, sn, el, en, link_id, raw_text, text))
                 break
-        el = sl + line_number
-        en = len(self.text_buffer[-1])
-        self.link_buffer.append((sl, sn, el, en, link_id, raw_text, text))
-        self.draw_last_line()
-        self.newline_required = newline_required
+
+    def append_meta(self, tag):
+        if not tag:
+            return
+        if not self.text_buffer:
+            sl = 0
+            sn = 0
+        else:
+            sl = len(self.text_buffer) - 1
+            sn = len(self.text_buffer[-1])
+        self.meta_buffer.append((sl, sn, tag))
 
     def append_image(self, path, x, y):
         try:
