@@ -12,7 +12,7 @@
 #  implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
 #  PURPOSE.  See the GNU General Public License for more details.
 #
-# $Id: sakura.py,v 1.148 2006/05/28 07:19:50 shy Exp $
+# $Id: sakura.py,v 1.150 2006/08/26 16:01:18 shy Exp $
 #
 
 import os
@@ -826,11 +826,12 @@ class Sakura:
         if self.anchor:
             anchor_id = self.anchor[0]
             text = self.anchor[1]
-            self.balloon.append_link(self.script_side, anchor_id, text)
+            self.balloon.append_link_out(self.script_side, anchor_id, text)
             self.anchor = None
         else:
             anchor_id = args[0]
             self.anchor = [('anchor', anchor_id), '']
+            self.balloon.append_link_in(self.script_side, self.anchor[0])
 
     def __yen_x(self, args):
         if self.script_mode == self.BROWSE_MODE:
@@ -997,45 +998,74 @@ class Sakura:
     def __yen_v(self, args):
         self.ghost.raise_surface(self.script_side)
 
+    def __yen_f(self, args):
+        if args[0] == 'sup':
+            assert len(args) == 2
+            if args[1] == 'true':
+                self.balloon.append_meta(self.script_side, '<sup>')
+            else:
+                self.balloon.append_meta(self.script_side, '</sup>')
+        elif args[0] == 'sub':
+            assert len(args) == 2
+            if args[1] == 'true':
+                self.balloon.append_meta(self.script_side, '<sub>')
+            else:
+                self.balloon.append_meta(self.script_side, '</sub>')
+        elif args[0] == 'strike':
+            assert len(args) == 2
+            if args[1] in ['true', '1', 1]:
+                self.balloon.append_meta(self.script_side, '<s>')
+            else:
+                self.balloon.append_meta(self.script_side, '</s>')
+        elif args[0] == 'underline':
+            assert len(args) == 2
+            if args[1] in ['true', '1', 1]:
+                self.balloon.append_meta(self.script_side, '<u>')
+            else:
+                self.balloon.append_meta(self.script_side, '</u>')
+        else:
+            pass ## FIXME
+
     __script_tag = {
-        r'\e': __yen_e ,
-        r'\0': __yen_0 ,
-        r'\h': __yen_0 ,
-        r'\1': __yen_1 ,
-        r'\u': __yen_1 ,
-        r'\p': __yen_p ,
-        r'\4': __yen_4 ,
-        r'\5': __yen_5 ,
-        r'\s': __yen_s ,
-        r'\b': __yen_b ,
-        r'\_b': __yen__b ,
-        r'\n': __yen_n ,
-        r'\c': __yen_c ,
-        r'\w': __yen_w ,
-        r'\_w': __yen__w ,
-        r'\t': __yen_t ,
-        r'\_q': __yen__q ,
-        r'\_s': __yen__s ,
-        r'\_e': __yen__e ,
-        r'\q': __yen_q ,
-        r'\URL': __yen_URL ,
-        r'\_a': __yen__a ,
-        r'\x': __yen_x ,
-        r'\a': __yen_a , # Obsolete: only for old SHIORI
-        r'\i': __yen_i ,
-        r'\j': __yen_j ,
-        r'\-': __yen_minus ,
-        r'\+': __yen_plus ,
-        r'\_+': __yen__plus ,
-        r'\m': __yen_m ,
-        r'\&': __yen_and ,
-        r'\_m': __yen__m ,
-        r'\_u': __yen__u ,
-        r'\_v': __yen__v ,
-        r'\!': __yen_exclamation ,
-        r'\__c': __yen___c ,
-        r'\__t': __yen___t , 
-        r'\v': __yen_v ,
+        r'\e': __yen_e,
+        r'\0': __yen_0,
+        r'\h': __yen_0,
+        r'\1': __yen_1,
+        r'\u': __yen_1,
+        r'\p': __yen_p,
+        r'\4': __yen_4,
+        r'\5': __yen_5,
+        r'\s': __yen_s,
+        r'\b': __yen_b,
+        r'\_b': __yen__b,
+        r'\n': __yen_n,
+        r'\c': __yen_c,
+        r'\w': __yen_w,
+        r'\_w': __yen__w,
+        r'\t': __yen_t,
+        r'\_q': __yen__q,
+        r'\_s': __yen__s,
+        r'\_e': __yen__e,
+        r'\q': __yen_q,
+        r'\URL': __yen_URL,
+        r'\_a': __yen__a,
+        r'\x': __yen_x,
+        r'\a': __yen_a, # Obsolete: only for old SHIORI
+        r'\i': __yen_i,
+        r'\j': __yen_j,
+        r'\-': __yen_minus,
+        r'\+': __yen_plus,
+        r'\_+': __yen__plus,
+        r'\m': __yen_m,
+        r'\&': __yen_and,
+        r'\_m': __yen__m,
+        r'\_u': __yen__u,
+        r'\_v': __yen__v,
+        r'\!': __yen_exclamation,
+        r'\__c': __yen___c,
+        r'\__t': __yen___t, 
+        r'\v': __yen_v,
+        r'\f': __yen_f,
         }
 
     def interpret_script(self):
@@ -1065,7 +1095,7 @@ class Sakura:
             text = self.expand_meta(node[1])
             if self.anchor:
                 self.anchor[1] = ''.join((self.anchor[1], text))
-            elif not self.quick_session and self.script_speed >= 0:
+            if not self.quick_session and self.script_speed >= 0:
                 self.processed_text = text
             else:
                 self.balloon.append_text(self.script_side, text)
