@@ -15,7 +15,6 @@
 
 import os
 import re
-import sys
 import urlparse
 import random
 
@@ -237,8 +236,7 @@ class Surface:
         # check if necessary surfaces have been loaded
         for key in ['0', '10']:
             if key not in pixbufs:
-                sys.stderr.write('cannot load surface #%s (abort)\n' % key)
-                sys.exit(1)
+                raise SystemExit, 'cannot load surface #%s (abort)\n' % key
         self.__pixbufs = pixbufs
         # arrange surface configurations
         region = {}
