@@ -10,8 +10,8 @@
 #  PURPOSE.  See the GNU General Public License for more details.
 #
 
-NUMBER = '3.8.9a'
-CODENAME = 'hacking run'
+NUMBER = '3.9'
+CODENAME = '"I see no more YASAGURE here."'
 
 VERSION = '%s (%s)' % (NUMBER, CODENAME)
 VERSION_INFO = (''.join((r'\h\s[0]\w8ninix-aya %s\n' % VERSION,
